/* serverpacket.c
 *
 * Constuct and send DHCP server packets
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>

#ifdef ACTION_TEC
	#include <unistd.h>
    #include <stdarg.h>
    #include <stdio.h>
    #include <stdlib.h>

#define ACTION_TEC_KPRINTF
#include <fcntl.h>
#define LED_KPRINTF 8
int kprintf (const char *fmt, ...)
{
#ifdef ACTION_TEC_KPRINTF
    static char tmp_text[51];
	va_list ap;
    int fp;

    /* Open /dev/led file for LED Operation */
    fp = open("/dev/led", O_RDWR);

    if(!fp) {
        return -1;
    }

    
    memset (tmp_text, 0, sizeof (tmp_text));
	va_start(ap, fmt);
	vsnprintf(tmp_text, sizeof (tmp_text)-1, fmt, ap);
    ioctl(fp, LED_KPRINTF, &tmp_text);

    close(fp);
    return 0;
#else
    return 0;
#endif //KPRINTF
}
#endif

#include "packet.h"
#include "debug.h"
#include "dhcpd.h"
#include "options.h"
#include "leases.h"

static int copy_till(char *inp , char *dest, char sep ,int length)
{
	int i = 0;
  while( (inp[i] != 0x00) && (inp[i] != sep) && (i < length ))
		*(dest + i) = inp[i++];
  *(dest + i) = 0x00;
  return i; 	
}

/* send a packet to giaddr using the kernel ip stack */
static int send_packet_to_relay(struct dhcpMessage *payload, int ifid)
{
	DEBUG(LOG_INFO, "Forwarding packet to relay");

	return kernel_packet(payload, server_config[ifid].server, SERVER_PORT,
			payload->giaddr, SERVER_PORT);
}


/* send a packet to a specific arp address and ip address by creating our own ip packet */
static int send_packet_to_client(struct dhcpMessage *payload, int force_broadcast, int ifid)
{
	unsigned char *chaddr;
	u_int32_t ciaddr;
	
	if (force_broadcast) {
		DEBUG(LOG_INFO, "broadcasting packet to client (NAK)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else if (payload->ciaddr) {
		DEBUG(LOG_INFO, "unicasting packet to client ciaddr");
		ciaddr = payload->ciaddr;
		chaddr = payload->chaddr;
	} else if (ntohs(payload->flags) & BROADCAST_FLAG) {
		DEBUG(LOG_INFO, "broadcasting packet to client (requested)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else {
		DEBUG(LOG_INFO, "unicasting packet to client yiaddr");
		ciaddr = payload->yiaddr;
		chaddr = payload->chaddr;
	}
	return raw_packet(payload, server_config[ifid].server, SERVER_PORT, 
			ciaddr, CLIENT_PORT, chaddr, server_config[ifid].ifindex);
}


/* send a dhcp packet, if force broadcast is set, the packet will be broadcast to the client */
static int send_packet(struct dhcpMessage *payload, int force_broadcast, int ifid)
{
	int ret;

	if (payload->giaddr)
		ret = send_packet_to_relay(payload, ifid);
	else ret = send_packet_to_client(payload, force_broadcast, ifid);
	return ret;
}


static void init_packet(struct dhcpMessage *packet, struct dhcpMessage *oldpacket, char type, int ifid)
{
	init_header(packet, type);
	packet->xid = oldpacket->xid;
	memcpy(packet->chaddr, oldpacket->chaddr, 16);
	packet->flags = oldpacket->flags;
	packet->giaddr = oldpacket->giaddr;
	packet->ciaddr = oldpacket->ciaddr;
	add_simple_option(packet->options, DHCP_SERVER_ID, server_config[ifid].server);
}


/* add in the bootp options */
static void add_bootp_options(struct dhcpMessage *packet, int ifid)
{
	packet->siaddr = server_config[ifid].siaddr;
	if (server_config[ifid].sname)
		strncpy(packet->sname, server_config[ifid].sname, sizeof(packet->sname) - 1);
	if (server_config[ifid].boot_file)
		strncpy(packet->file, server_config[ifid].boot_file, sizeof(packet->file) - 1);
}
	

/* send a DHCP OFFER to a DHCP DISCOVER */
int sendOffer(struct dhcpMessage *oldpacket, int ifid)
{
	struct dhcpMessage packet;
	struct dhcpOfferedAddr *lease = NULL;
	u_int32_t req_align, lease_time_align = server_config[ifid].lease,leasetime = 0;
	unsigned char *req, *lease_time;
	struct option_set *curr;
	struct in_addr addr;
  u_int8_t hostname[50] = "unknown" , *hname;
  int length = 0;

	init_packet(&packet, oldpacket, DHCPOFFER, ifid);
	
	/* ADDME: if static, short circuit */
	/* the client is in our lease/offered table */
	if ((lease = find_lease_by_chaddr(oldpacket->chaddr, ifid)))
  {
		if (!lease_expired(lease,ifid))
		{
			if(lease->expires == server_config[ifid].inflease_time)
				lease_time_align = server_config[ifid].inflease_time;
      else
				lease_time_align = lease->expires - time(0);
		}
		packet.yiaddr = lease->yiaddr;
		
	/* Or the client has a requested ip */
	} else if ((req = get_option(oldpacket, DHCP_REQUESTED_IP)) &&

		   /* Don't look here (ugly hackish thing to do) */
		   memcpy(&req_align, req, 4) &&

		   /* and the ip is in the lease range */
		   ntohl(req_align) >= ntohl(server_config[ifid].start) &&
		   ntohl(req_align) <= ntohl(server_config[ifid].end) &&
		   
		   /* and its not already taken/offered */ /* ADDME: check that its not a static lease */
		   ((!(lease = find_lease_by_yiaddr(req_align, ifid)) ||
		   
		   /* or its taken, but expired */ /* ADDME: or maybe in here */
		   lease_expired(lease,ifid)))) {
		   /* check id addr is not taken by a static ip */
		   if(!check_ip(req_align, ifid)) 	
				packet.yiaddr = req_align; /* FIXME: oh my, is there a host using this IP? */
		   else {
			   packet.yiaddr = find_address(0, ifid);

			   /* try for an expired lease */
			   if (!packet.yiaddr) packet.yiaddr = find_address(1, ifid);
			}

			

	/* otherwise, find a free IP */ /*ADDME: is it a static lease? */
	}
  else
  {
		packet.yiaddr = find_address(0, ifid);
		/* try for an expired lease */
		if (!packet.yiaddr) packet.yiaddr = find_address(1, ifid);
	}
	
	if(!packet.yiaddr) {
		LOG(LOG_WARNING, "no IP addresses to give -- OFFER abandoned");
		return -1;
	}

  hname = get_option(oldpacket,DHCP_HOST_NAME);
  if(hname)
  {
    add_option_string(packet.options, hname - 2);
		memset(hostname,0x00,50);
    length = get_option_length(oldpacket,DHCP_HOST_NAME);
    copy_till(hname , hostname, '.',length);
		LOG(LOG_INFO, "SENDING OFFER to %s\n",hostname);
  }	

  /*Check for infinite lease */
	if ((lease = find_lease_by_chaddr(oldpacket->chaddr, ifid)))
	{
		if(lease->expires == server_config[ifid].inflease_time)
			leasetime = server_config[ifid].inflease_time;
		else
				leasetime = server_config[ifid].offer_time;
	}
  else
		leasetime = server_config[ifid].offer_time;

	if (!add_lease(packet.chaddr, packet.yiaddr, leasetime /*server_config[ifid].offer_time*/, ifid,hostname)) {
		LOG(LOG_WARNING, "lease pool is full -- OFFER abandoned");
		return -1;
	}		

	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME))) {
		memcpy(&lease_time_align, lease_time, 4);
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config[ifid].lease) 
			lease_time_align = server_config[ifid].lease;
	}

	/* Make sure we aren't just using the lease time from the previous offer */
	if (lease_time_align < server_config[ifid].min_lease) 
		lease_time_align = server_config[ifid].lease;

  /* For inifinite leases change the lease time */
  if( leasetime == server_config[ifid].inflease_time)
    lease_time_align = leasetime;
	/* ADDME: end of short circuit */		
	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	curr = server_config[ifid].options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet, ifid);
	
	addr.s_addr = packet.yiaddr;
	LOG(LOG_INFO, "sending OFFER of %s", inet_ntoa(addr));

	return send_packet(&packet, 0, ifid);
}


int sendNAK(struct dhcpMessage *oldpacket, int ifid)
{
	struct dhcpMessage packet;

	init_packet(&packet, oldpacket, DHCPNAK, ifid);
	
	DEBUG(LOG_INFO, "sending NAK");
	return send_packet(&packet, 1, ifid);
}


int sendACK(struct dhcpMessage *oldpacket, u_int32_t yiaddr, int ifid)
{
	struct dhcpMessage packet;
	struct option_set *curr;
	unsigned char *lease_time;
	u_int32_t lease_time_align = server_config[ifid].lease;
	struct in_addr addr;
  u_int8_t hostname[50]="" , *hname;
  int length = 0;
	struct dhcpOfferedAddr *lease = NULL;
#ifdef ACTION_TEC
    char line[18]={0};
    int fp_flag=0;
#endif
	init_packet(&packet, oldpacket, DHCPACK, ifid);
	packet.yiaddr = yiaddr;
	
	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME))) {
		memcpy(&lease_time_align, lease_time, 4);
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config[ifid].lease) 
			lease_time_align = server_config[ifid].lease;
		else if (lease_time_align < server_config[ifid].min_lease) 
			lease_time_align = server_config[ifid].lease;
	}
	
  /* If the existing lease entry has infinite entry give it infinite time */
	if ( (lease = find_lease_by_chaddr(oldpacket->chaddr, ifid)) )
	{
		if(lease->expires == server_config[ifid].inflease_time)
			lease_time_align = server_config[ifid].inflease_time;
	}

	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

  hname = get_option(oldpacket,DHCP_HOST_NAME);
  if(hname)
  {
    add_option_string(packet.options, hname - 2 );
    memset(hostname,0x00,50);
    length = get_option_length(oldpacket,DHCP_HOST_NAME);
    copy_till(hname , hostname, '.',length);
		LOG(LOG_INFO, "SENDING ACK to %s\n",hostname);
  }	
	
	curr = server_config[ifid].options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet, ifid);

	addr.s_addr = packet.yiaddr;
	LOG(LOG_INFO, "sending ACK to %s", inet_ntoa(addr));

	if (send_packet(&packet, 0, ifid) < 0) 
		return -1;

	add_lease(packet.chaddr, packet.yiaddr, lease_time_align, ifid,hostname);
#ifdef ACTION_TEC_VERIZON_DEMO
    { //char line[17]={0}; 
      sprintf(line,"%02x-%02x-%02x-%02x-%02x-%02x",
      packet.chaddr[0],packet.chaddr[1],packet.chaddr[2],packet.chaddr[3],packet.chaddr[4],packet.chaddr[5]);
    }
#endif

#ifdef ACTION_TEC_DYNLINK
    {//when renew if firewall is not ready, this file will be used when firewall start (fwan)
        #define FILE_HOST_RENEW "/var/tmp/dynlink/hosts_renew"
        char ip_peer[20];
        char buf_redirect[200];
        struct in_addr saddr;
        int exist = 0;
        FILE *fp = fopen (FILE_HOST_RENEW, "r");

        saddr.s_addr = packet.yiaddr;
        strcpy (ip_peer, inet_ntoa (saddr));
        sprintf (buf_redirect, 
                 "/sbin/iptables -D FORWARD -s %s -p tcp ! --syn --dport 80 -j QUEUE",
                 ip_peer);
        system (buf_redirect);
        sprintf (buf_redirect, 
                 "/sbin/iptables -I FORWARD -s %s -p tcp ! --syn --dport 80 -j QUEUE &",
                 ip_peer);
        system (buf_redirect);

        if (fp) {
            char ipbuf[20];
            memset (ipbuf, 0, sizeof (ipbuf));
            while (fgets (ipbuf,sizeof (ipbuf)-1, fp)) {
                if (ipbuf[strlen (ipbuf)-1] == '\n') {
                    ipbuf[strlen (ipbuf)-1] = 0;
                }
                if (!strcmp (ip_peer, ipbuf)) {
                    exist = 1;
                    break;
                }
                memset (ipbuf, 0, sizeof (ipbuf));
            }
            fclose (fp);
        }
        if (!exist) {
            sprintf (buf_redirect, "echo \"%s\" >> %s&",ip_peer, FILE_HOST_RENEW);
            system (buf_redirect);
        }
    }
#endif //ACTION_TEC_DYNLINK
#ifdef ACTION_TEC
	{
		struct in_addr ip_addr;
		unsigned char *hostname, hosts[256]={0};

		hostname = get_option(oldpacket, DHCP_HOST_NAME);
		if( hostname ) 
			strncpy(hosts, hostname, hostname[OPT_LEN - 2]);
		else
			strcpy(hosts,  "unknown");
		ip_addr.s_addr = packet.yiaddr;

		if ( strlen(hosts) && yiaddr!=0 ) {
			char buf[510];
			FILE* fp1 = fopen("/var/tmp/hosts.dhcp", "r");
			FILE* fp2 = fopen("/var/tmp/hosts.2", "w");
			if ( fp2 ) {
				int iFound;
				unsigned char *ipstr,tmpstr[25];

				iFound = 0;
				//ipstr = inet_ntoa(ip_addr);
				strncpy(tmpstr,inet_ntoa(ip_addr),25);
				ipstr=&tmpstr[0];
				if ( fp1 ) {
					while ( fgets(buf, 510, fp1) ) {
						char *s, *s1;
						s = strtok(buf, " \r\n\t");
						if ( s ) {
							if ( strcmp(s, ipstr)==0 ) {
								iFound = 1;    
								fprintf(fp2, "%s\t%s\n", ipstr, hosts);
							} else {
								s1 = strtok(NULL, "\r\n");
								fprintf(fp2, "%s\t%s\n", s, s1); 
							}   
						}
					}   
				}
				if ( !iFound )
					fprintf(fp2, "%s\t%s\n", ipstr, hosts);
			};
			if ( fp1 ) {
				fclose(fp1);
				unlink("/var/tmp/hosts.dhcp");
			}
			if ( fp2 ) {
				fclose(fp2);
				rename("/var/tmp/hosts.2", "/var/tmp/hosts.dhcp");
				//system("/sbin/cli_query.sh tr69 host_num > /var/tmp/t &");
			}
		}
	}       
#endif	
#ifdef ACTION_TEC   //for ALG port forwarding
    {
        FILE *fp, *fp_bk;
        char *ip, *pcname;
        int ip_changed = 0;
        char buf[100], buf_bk[100];
        char ip_offer[20];

        //user fp_bk to prevent from sending cli cmd twice for repeat offer
        strcpy (ip_offer, inet_ntoa(addr));
        if (((fp = fopen ("/var/tmp/dhcp_iphost", "r")) != NULL)){
            if ((fp_bk = fopen ("/var/tmp/dhcp_iphost.bk", "w")) != NULL) {
                sprintf (buf, "%s %s\n", ip_offer, hostname);
                fwrite (buf, sizeof (char), strlen (buf),fp_bk);    //keep the newest one on the top
                                            //in case one pc has 2 connections
                while ((fgets (buf, sizeof (buf)- 1, fp)) != NULL) {
                    fp_flag=1;
                    memcpy (buf_bk, buf, sizeof (buf_bk));
                    buf_bk[sizeof (buf_bk) -1] = 0;
                    ip = strtok (buf, " \t");
                    pcname= strtok (NULL, " \t\n");
                    if (ip && pcname && (strcmp (pcname, hostname) == 0) ) {
                        if (!ip_changed ) {
                            if (strcmp (ip, ip_offer) == 0) {//already exist and is the first one
                                 break;
                            }
                            else {
                                system ("cp -f /var/tmp/dhcp_iphost /var/tmp/dhcp_iphost.old");
                                ip_changed = 1;
                            }
                        }
                    }
                    else {
                        fwrite (buf_bk, sizeof (char), strlen (buf_bk), fp_bk);
                    }
                    memset (buf, 0, sizeof (buf));
                }
                fclose (fp_bk);
#ifdef ACTION_TEC_VERIZON_DEMO
//if no content in dhcp_iphost, we won't take it as ip changed any more
//                if (fp_flag==0) {
//                    ip_changed = 1;
//                }
                system ("cp -f /var/tmp/dhcp_iphost /var/tmp/dhcp_iphost.old");
                system ("cp -f /var/tmp/dhcp_iphost.bk /var/tmp/dhcp_iphost");
#else
                if (ip_changed) {
                    system ("cp -f /var/tmp/dhcp_iphost.bk /var/tmp/dhcp_iphost");
                }
#endif
            }
            fclose (fp);
        }
//        kprintf ("--%d send ack=%s,%s\n", ip_changed, hostname, ip_offer);
        if (ip_changed) {
#ifdef ACTION_TEC_VERIZON_DEMO
            //double_command.sh has influence on Verizon WBA application, use cli_query instead
//            sprintf (buf, "double_command.sh connection0 fwan %s/%s change_ip 4&",
//                     line, ip_offer);
            sprintf (buf, "/sbin/cli_query.sh connection0 fwan:settings/change_ip/%s/%s 1 2> /dev/null",
                        line, ip_offer);
            kprintf("\nchanged ip: %s %s\n",line,ip_offer);
#else
            sprintf (buf, "double_command.sh connection0 fwan %s/%s change_ip 4&",
                     hostname, ip_offer);
#endif
            system (buf);
        }
#ifdef ACTION_TEC_REMOTE_SYSLOG   
    {
        FILE *fp;
        char *ip, *port_str, *pcname;
        char buf[100];
        int ip_changed = 0;
        char ip_offer[20];

        //user fp_bk to prevent from sending cli cmd twice for repeat offer
        strcpy (ip_offer, inet_ntoa(addr));
        memset (buf, 0, sizeof (buf));
        if (((fp = fopen ("/var/tmp/remote_syslog_server", "r")) != NULL)){
            if ((fgets (buf, sizeof (buf)- 1, fp)) != NULL) {
                ip = strtok (buf, " \t\n");
                port_str = strtok (NULL, " \t\n");
                pcname= strtok (NULL, " \t\n");
//                kprintf ("ack=%s,%s,%s\n", hostname, port_str, pcname);
                if (ip && pcname && (strcmp (pcname, hostname) == 0) ) {
                    if (strcmp (ip, ip_offer)) {//different from the new one
                        ip_changed = 1;
                    }
                }
            }
            fclose (fp);
        }
//        kprintf ("%d n ack=%s,%s\n", ip_changed, hostname, ip_offer);
        if (ip_changed) {
            sprintf (buf, "/sbin/single_settings.sh security server_remote_syslog %s &", ip_offer);
            system (buf);
        }
    }
#endif ACTION_TEC_REMOTE_SYSLOG
    }
#endif ACTION_TEC

	return 0;
}


int send_inform(struct dhcpMessage *oldpacket, int ifid)
{
	struct dhcpMessage packet;
	struct option_set *curr;

	init_packet(&packet, oldpacket, DHCPACK, ifid);
	
	curr = server_config[ifid].options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet, ifid);

	return send_packet(&packet, 0, ifid);
}



